#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


/********************************************************************/
/****************************** LWL93 *******************************/
/********************************************************************/

/* Le programme lwl93 calcule les taus de substitutions synonymes */
/* et non-synonymes au sens de Li (1993, J.M.E.36. 96-99) entre */
/* toutes les paires de sequences d'un fichier mase. La sortie est */
/* un ou deux fichier .num contenant les Ka et/ou Ks avec/sans leurs */
/* variances. */


/* code_mt TRUE ssi code genetique mt vertebre
sinon FALSE pour code genetique universel */
int code_mt;


main()
{
#define maxnseqs 100
	double          *tl0[64], *tl1[64], *tl2[64], *tti0[64], *tti1[64],
	               *tti2[64], *ttv0[64], *ttv1[64], *ttv2[64];
	char           *seq[maxnseqs], *seqname[maxnseqs], masefname[40], outfname[40];
	char           *comments[maxnseqs], fname[100], opt[12], option3[3],tamp[2];
	int             i, j, j3, l, totseqs, lgseq, n, option2;
	double          *ka[100], *ks[100], *rl[21], *vka[100], *vks[100];
	FILE           *output;

	int             readmaseseqs(char *, char **, char **, char **, int);
	void            refresh(char **, int, int);

	void            prefastlwl(double **, double **, double **, double **, double **, double **, double **, double **, double **, double **);
	void            fastlwl(char **, int, int, double **, double **, double **, double **, double **, double **, double **, double **, double **, double **, double **, double **, double **);
char *check_alloc(int, int);

  for (i = 0; i < 64; i++) {
    tl0[i] = (double *) check_alloc(64, sizeof(double));
    tl1[i] = (double *) check_alloc(64, sizeof(double));
    tl2[i] = (double *) check_alloc(64, sizeof(double));
    tti0[i] = (double *) check_alloc(64, sizeof(double));
    tti1[i] = (double *) check_alloc(64, sizeof(double));
    tti2[i] = (double *) check_alloc(64, sizeof(double));
    ttv0[i] = (double *) check_alloc(64, sizeof(double));
    ttv1[i] = (double *) check_alloc(64, sizeof(double));
    ttv2[i] = (double *) check_alloc(64, sizeof(double));
  }

  for (i = 0; i < 21; i++) rl[i] = (double *) check_alloc(21,sizeof(double));
  for (i = 0; i < 21; i++) rl[0][i]=rl[i][0]=0.;


  rl[2][1]=rl[3][1]=rl[3][2]=rl[5][1]=rl[5][3]
          =rl[6][1]=rl[6][3]=rl[6][5]=0.382;
  rl[7][1]=rl[7][3]=rl[7][5]=rl[7][6]=rl[8][1]
          =rl[8][5]=rl[8][6]=rl[8][7]=0.382;
  rl[11][9]=rl[12][9]=rl[13][4]=rl[13][9]=rl[15][4]
           =rl[15][13]=rl[16][14]=rl[16][15]=0.382;
  rl[17][4]=rl[18][4]=rl[18][15]=rl[18][17]=rl[19][4]
           =rl[19][15]=rl[19][16]=rl[20][16]=0.382;
  rl[20][19]=0.382;

  rl[4][1]=rl[4][3]=rl[5][2]=rl[5][4]=rl[6][2]
          =rl[6][4]=rl[7][2]=rl[7][4]=0.343;
  rl[8][2]=rl[8][3]=rl[8][4]=rl[9][4]=rl[9][5]
          =rl[9][6]=rl[9][7]=rl[9][8]=0.343;
  rl[11][4]=rl[11][5]=rl[11][6]=rl[11][7]=rl[11][8]
           =rl[12][4]=rl[12][11]=rl[13][3]=0.343;
  rl[13][5]=rl[13][6]=rl[13][7]=rl[13][8]=rl[13][11]
           =rl[13][12]=rl[14][4]=rl[14][9]=0.343;
  rl[14][11]=rl[14][12]=rl[14][13]=rl[15][3]=rl[15][8]
            =rl[15][9]=rl[15][11]=rl[15][12]=0.343;
  rl[15][14]=rl[16][4]=rl[16][9]=rl[16][12]=rl[16][13]
            =rl[17][3]=rl[17][5]=rl[17][8]=0.343;
  rl[17][13]=rl[17][15]=rl[17][16]=rl[18][1]=rl[18][3]
            =rl[18][5]=rl[18][7]=rl[18][8]=0.343;
  rl[18][13]=rl[18][16]=rl[19][9]=rl[19][12]=rl[19][13]
            =rl[19][14]=rl[19][17]=rl[19][18]=0.343;
  rl[20][4]=rl[20][12]=rl[20][13]=rl[20][14]=rl[20][15]
           =rl[20][18]=0.343;

  rl[4][2]=rl[9][1]=rl[9][2]=rl[9][3]=rl[11][1]
          =rl[11][2]=rl[11][3]=rl[12][5]=0.128;
  rl[12][6]=rl[12][7]=rl[12][8]=rl[13][1]=rl[13][2]
           =rl[13][10]=rl[14][3]=rl[14][5]=0.128;
  rl[14][6]=rl[14][7]=rl[14][8]=rl[14][10]=rl[15][1]
           =rl[15][2]=rl[15][5]=rl[15][6]=0.128;
  rl[15][7]=rl[16][3]=rl[16][5]=rl[16][7]=rl[16][8]
           =rl[16][10]=rl[16][11]=rl[17][1]=0.128;
  rl[17][2]=rl[17][6]=rl[17][7]=rl[17][9]=rl[17][11]
           =rl[17][12]=rl[17][14]=rl[18][2]=0.128;
  rl[18][6]=rl[18][9]=rl[18][11]=rl[18][12]=rl[18][14]
           =rl[19][1]=rl[19][3]=rl[19][5]=0.128;
  rl[19][6]=rl[19][7]=rl[19][8]=rl[19][11]=rl[20][9]
       =rl[20][11]=rl[20][17]=rl[12][3]=0.128;

  rl[10][1]=rl[10][2]=rl[10][3]=rl[10][4]=rl[10][5]
           =rl[10][6]=rl[10][7]=rl[10][8]=0.040;
  rl[10][9]=rl[11][10]=rl[12][1]=rl[12][2]=rl[12][10]
           =rl[14][1]=rl[14][2]=rl[15][10]=0.040;
  rl[16][1]=rl[16][2]=rl[16][6]=rl[17][10]=rl[18][10]
           =rl[19][2]=rl[19][10]=rl[20][1]=0.040;
  rl[20][2]=rl[20][3]=rl[20][5]=rl[20][6]=rl[20][7]
           =rl[20][8]=rl[20][10]=0.040;

  for (i = 1; i <= 20; i++) {
    *(rl[i] + i) = 1.0;
    for (j = i + 1; j <= 20; j++)
      *(rl[i] + j) = *(rl[j] + i);
  }


  printf("\nThis program computes synonymous and non-synonymous substitutions between protein coding genes according to the method of Li, Wu and Luo (Mol. Biol. Evol. 2, 1985) modified by Li (J. Mol. Evol. 1993).\n\n");

  printf("Name of file of aligned sequences? (mase)\n");
  gets(masefname);
  totseqs = readmaseseqs(masefname, seq, seqname, comments, maxnseqs);
  lgseq = strlen(seq[0]);

  printf("\nGenetic code? Universal(u) or Mitochondrial(m) ");
  gets(fname);
  code_mt = (*fname=='m' || *fname=='M');
	
  for (i=0;i<totseqs;i++){
    for(j=0;j<lgseq;j++){
      if ((*(seq[i]+j)!='A') && (*(seq[i]+j)!='G') && (*(seq[i]+j)!='C') && (*(seq[i]+j)!='T')) {
	j3=j%3;
	seq[i][j-j3]=seq[i][j+1-j3]=seq[i][j+2-j3]='-';
      }
    }
  }


  refresh(seq, totseqs, 0);

  for (i = 0; i <= totseqs; i++) {
    ka[i] = (double *) check_alloc(totseqs + 1, sizeof(double));
    vka[i] = (double *) check_alloc(totseqs + 1, sizeof(double));
    ks[i] = (double *) check_alloc(totseqs + 1, sizeof(double));
    vks[i] = (double *) check_alloc(totseqs + 1, sizeof(double));
  }

  lgseq = strlen(seq[0]);

  prefastlwl(rl, tl0, tl1, tl2, tti0, tti1, tti2, ttv0, ttv1, ttv2);
  fastlwl(seq, totseqs, lgseq, ka, ks, tti0, tti1, tti2, ttv0, ttv1, ttv2, tl0, tl1, tl2, vka, vks);

  if(totseqs<=10){

    printf("Ks:\n");
    for (i = 0; i < totseqs - 1; i++) {
      if (i != 0)
	printf("\n");
      for (j = i + 1; j < totseqs; j++) {
	printf("%f  ", *(ks[i] + j));
      }
    }
    printf("\n\nKa:\n");
    for (i = 0; i < totseqs - 1; i++) {
      if (i != 0) printf("\n");
      for (j = i + 1; j < totseqs; j++) 
	printf("%f  ", *(ka[i] + j));
    }
    printf("\n\nV(Ks):\n");
    for (i = 0; i < totseqs - 1; i++) {
      if (i != 0) printf("\n");
      for (j = i + 1; j < totseqs; j++) 
	printf("%f  ", *(vks[i] + j));
    }
    printf("\n\nV(Ka):\n");
    for (i = 0; i < totseqs - 1; i++) {
      if (i != 0) printf("\n");
      for (j = i + 1; j < totseqs; j++) 
	printf("%f  ", *(vka[i] + j));
    }
    printf("\n\n");
    printf("(9.999999 = saturation)\n");
  }

  printf("\nOutput :");
  if(totseqs<=10) printf("  - nothing (0)\n"); else printf("\n");
  printf("          - .num file (=input to nj) of synonymous rates (1)\n");
  printf("          - .num file (=input to nj) of non-synonymous rates (2)\n");
  printf("          - both files (3)\n");
  gets(tamp);
  sscanf(tamp,"%d", &option2);

  if (option2 == 0) exit(0);
  printf("\nDo you want variances? (y/[n])\n");

  gets(option3);
  if (option3[0]==0) option3[0]='y';

  if ((option2 == 1) || (option2 == 3)) {
    printf("\nName of the ");
    if (option2 == 3) printf("1st ");
    printf("output file ");
    if (option2 == 3) printf("(Ks) ");
    printf("? [use .num extension for name]\n");
    gets(outfname);

    strcpy(opt, "Option: Ks\n");

    output = fopen(outfname, "w");
    fprintf(output, "%d\n", totseqs);
    for (n = 0; n < totseqs - 1; n++) {
      j = 0;
      for (i = 0; i < totseqs; i++) {
	if (n < i) {
	  fprintf(output, "   %.6f", *(ks[n] + i));
	  j++;
	  if (j % 7 == 0) fprintf(output, "\n");
	}
      }
      if (j % 7 != 0) fprintf(output, "\n");

    }
    j=strlen(opt);
    for (i = 0; i < j; i++) fprintf(output, "%c", *(opt + i));
    for (i = 0; i < totseqs; i++) {
      fprintf(output, "%s", seqname[i]);
      l = 20 - strlen(seqname[i]);
      for (j = 0; j < l; j++) fprintf(output, " ");
      if ((i + 1) % 4 == 0) fprintf(output, "\n");
    }

    fprintf(output, "\n");


    if (option3[0] == 'y') {
      fprintf(output, "\n");
      fprintf(output, "Variances:\n");

      for (n = 0; n < totseqs - 1; n++) {
	j = 0;
	for (i = 0; i < totseqs; i++) {
	  if (n < i) {
	    fprintf(output, "   %.6f", *(vks[n] + i));
	    j++;
	    if (j % 7 == 0) fprintf(output, "\n");
	  }
	}
	if (j % 7 != 0)	fprintf(output, "\n");

      }
    }
    fclose(output);
  }
	
  if ((option2 == 2) || (option2 == 3)) {
    printf("Name of the ");
    if (option2 == 3) printf("2nd ");
    printf("output file ");
    if (option2 == 3) printf("(Ka) ");
    printf("? [use .num extension for name]\n");
    gets(outfname);

    strcpy(opt, "Option: Ka\n");

    output = fopen(outfname, "w");
    fprintf(output, "%d\n", totseqs);
    for (n = 0; n < totseqs - 1; n++) {
      j = 0;
      for (i = 0; i < totseqs; i++) {
	if (n < i) {
	  fprintf(output, "   %.6f", *(ka[n] + i));
	  j++;
	  if (j % 7 == 0) fprintf(output, "\n");
	}
      }
      if (j % 7 != 0) fprintf(output, "\n");

    }
    j=strlen(opt);
    for (i = 0; i < j; i++) fprintf(output, "%c", *(opt + i));
    for (i = 0; i < totseqs; i++) {
      fprintf(output, "%s", seqname[i]);
      l = 20 - strlen(seqname[i]);
      for (j = 0; j < l; j++) fprintf(output, " ");
      if ((i + 1) % 4 == 0) fprintf(output, "\n");
    }

    fprintf(output, "\n");
	
    if (option3[0]=='y'){

      fprintf(output, "\n");
      fprintf(output, "Variances:\n");
      for (n = 0; n < totseqs - 1; n++) {
        j = 0;
        for (i = 0; i < totseqs; i++) {
	  if (n < i) {
	    fprintf(output, "   %.6f", *(vka[n] + i));
	    j++;
	    if (j % 7 == 0) fprintf(output, "\n");
	  }
        }
        if (j % 7 != 0) fprintf(output, "\n");

      }

    }
    close(output);
  }
}
